import java.io.*; 
import java.util.*;
import java.nio.*;
import java.nio.channels.*;
import javax.imageio.stream.*;
import javax.swing.*;
import java.awt.*;
import java.util.regex.*;

import cj.appkit.*;
import cj.math.jama.*;
import cj.fileformats.*;
import gnu.getopt.*;

/*
 *  The idea of this class is to read in the whole gradient
 *  dataset from the DPF file or a text file and to store
 *  the information internally.
 *
 *  All grdaients will be stored here as it will relate
 *  to a specific REC file.  Therefore, a vector of (0,0,0)
 *  will represent a b=0 volume, a vector of (100,100,100) 
 *  will represent a trace volume.
 */
public class Gradients
{
	//  Filename
	private String filename = null;

	// Verbosity
	private int verbose = 0;

	//  b value used...
	// private double b = 700.0;  // default

	//  This will be a list of all the gradients, 
	//  including the (0,0,0) and (100,100,100)
	private Matrix gradients = null;

	private int[] reference_indices = null;
	private int[] direction_indices = null;
	
	public Gradients()
	{
	}

	public Gradients(String filename)
	{
		readGradients( filename );
	}

	//  This will return all the gradients which are not in
	//  the removeGradientsTF vector.
	final public Matrix getGradients()
	{
		return gradients;
	}

	final public int[] getReferenceIndices()
	{
		return reference_indices;
	}

	final public int[] getDirectionIndices()
	{
		return direction_indices;
	}

	/**
		* Read in the gradient table from a dpf file
		* @param filename
		* @ return gradients
		*/
	public void readGradients(String filename)
	{
		Pattern pattern = Pattern.compile("\\s*\\d+:\\s*([\\-0-9\\.]*)\\s*,\\s*([\\-0-9\\.]*)\\s*,\\s*([\\-0-9\\.]*).*");

		// Create temporary bins for the Gradient information
		ArrayList<Double> gx_vec = new ArrayList<Double>(30);
		ArrayList<Double> gy_vec = new ArrayList<Double>(30);
		ArrayList<Double> gz_vec = new ArrayList<Double>(30);

		ArrayList<Integer> ref_arr = new ArrayList<Integer>(30);
		ArrayList<Integer> dir_arr = new ArrayList<Integer>(30);

		//-------------------------------------------------------
		//
		//  Read in the gradient directions.
		//
		//-------------------------------------------------------
		ImageInputStream ldis = AppKit.openRead(filename);

		String line = new String();

		boolean done = false;
		boolean found_begin = false;
		do
		{
			try {
				line = ldis.readLine();	

				if( line == null ) { continue; }
				
				Matcher matcher = pattern.matcher(line);
				if( matcher.matches() )
				{
					gx_vec.add( new Double(matcher.group(1)) );
					gy_vec.add( new Double(matcher.group(2)) );
					gz_vec.add( new Double(matcher.group(3)) );
				}
			}
			catch(EOFException e) { done = true; }
			catch(IOException e) { done = true; }
		}
		while( !done && line != null );

		//----------------------------------------------------------
		//
		// Now copy over the gradient info from the vectors
		// to the matrix which will then be returned.
		//
		//----------------------------------------------------------
		gradients = new Matrix(gx_vec.size(), 3);
		for(int ii=0; ii<gx_vec.size(); ii++) 
		{
			double gx = ((Double)gx_vec.get(ii)).doubleValue();
			double gy = ((Double)gy_vec.get(ii)).doubleValue();
			double gz = ((Double)gz_vec.get(ii)).doubleValue();

			gradients.set(ii,0, gx);
			gradients.set(ii,1, gy);
			gradients.set(ii,2, gz);

			if( ( Math.abs(gx) + Math.abs(gy) + Math.abs(gz)) == 0 )
				ref_arr.add(ii);
			else if( ( Math.abs(gx) + Math.abs(gy) + Math.abs(gz)) < 4 )
				dir_arr.add(ii);
		}

		// Copy over the refeerence and direction indices
		reference_indices = new int[ ref_arr.size() ];
		for(int ii=0; ii<reference_indices.length; ii++) reference_indices[ii] = ref_arr.get(ii);

		direction_indices = new int[ dir_arr.size() ];
		for(int ii=0; ii<direction_indices.length; ii++) direction_indices[ii] = dir_arr.get(ii);

		if( verbose > 2 )
		{ 
			System.out.println("Gradient table is " + gradients.getRowDimension() +" by " + gradients.getColumnDimension());
			gradients.print(3,3); 
		}
	}	

}
