package apps.drifter;

import cj.appkit.*;
import cj.volume.*;
import cj.volume.measure.*;
import cj.volume.process.*;
import cj.utilities.*;
import cj.fileformats.*;
import cj.math.matrix.*;
import cj.math.discrete.*;

/**
 * DrifterApp - measures short/long discrepancy
 */

public class DrifterApp {
	static public void main(String[] argv) { new DrifterKit().main(argv); }
}

class DrifterKit extends AppKit
{
    static final int minArea = 4000;
    int skippedSlices = 0;
	int radius = 10;
	int fix_level = 0;
	int hits=0, miss=0, total=0;
    int[] maskArea;
	double waste=0;

	MCVolume v; Volume maskv;
	MRIFile mr; MIFName mn;

	BinaryVolume mask = null;

	public String getAppName() { return "drifter"; }
	public String getDialogName() { return "Select MIF pair"; }

	protected void moreOpts()
	{
		addOpt_reqArg("radius",'r',"maximum pixel displacement to scan for");
		addOpt_noArg("fix",'f',"output fixed long-echo");
	}

	public boolean parseArg(int c)
	{
		switch(c)
		{
			case 'r':
				radius = getIntArg("radius"); break;
			case 'f':
				fix_level++; break;
			default:
				return false;
		}
		return true;
	}

	public void help()
	{
		System.out.println("Usage: "+getAppName()+" input.MIF [mask.MIF]\n");
		printOptions();
		System.out.println("mask must have the same dims as input and contain only 0/1 values.");
	}

	private void loadData()
	{
		if (otherArgs.length > 2) { help(); System.exit(-1); }

		// Load mask, if specified
		if (otherArgs.length == 2)
		{
			mr = MRIFile.readMRIFile(otherArgs[1]);
			if (mr == null)
			{
				System.err.println("Could not open mask " + otherArgs[1]);
				System.exit(-1);
			}
			maskv = mr.getMCVolume().getVolume(0);
			mask = maskv.threshold(0.5,1.5);
		}

		mn = new MIFName(otherArgs[0]);
		String file;

		// Now load short-echo
		file = mn.recode('S').toString();
		mr = MRIFile.readMRIFile(file);

		if (mr == null)
		{
			System.err.println("Could not open volume " + file);
			System.exit(-1);
		}
		v = mr.getMCVolume();

		// Then long-echo
		file = mn.recode('L').toString();
		mr = MRIFile.readMRIFile(file);

		if (mr == null)
		{
			System.err.println("Could not open volume " + file);
			System.exit(-1);
		}

		// v is guaranteed to be short-echo first
		v.concatenate(mr.getMCVolume());

		// On exit, mr and mn are still long-echo.
	}

	/**
	 * Computes mutual information for every offset on every slice.
	 */

	private double[][] compileTable(int radius)
	{
		//Affine af = new Affine();
		//af.setInterpolationMethod(Affine.NEAREST_NEIGHBOR_INTERPOLATION);
		//af.setAffineMethod(Affine.AFFINE_ROTATE);

		Slice s1, s2; BinarySlice sm=null;
		int i,j,k;
		int n = 2*radius+1, m = v.getNSlices();

		double[][] tab = new double[m][n];
		double[] lb={0,0}, ub=new double[2], bins = new double[2];

		// Default weight is entire slice.
		double weight = v.getNRows()*v.getNCols();

		maskArea = new int[m];

		for(i=0;i<m;i++)
		{
			s1 = v.getVolume(0).getSlice(i);
			s2 = v.getVolume(1).getSlice(i);

			// Weight entropy by area of mask (total information versus
			// information per pixel).

			if (mask != null) { sm=mask.getSlice(i); weight = Statistics.sum(maskv.getSlice(i)); }

			if (mask == null) { maskArea[i] = s1.getNRows()*s1.getNCols(); }
            else { maskArea[i] = (int)weight; }

			ShiftSlice sr = new ShiftSlice(s2);

			// Set histogram parameters.  Fixed step-size of 10.
			ub[0] = Statistics.max(s1); ub[1] = Statistics.max(s2);
			bins[0] = ub[0]/10; bins[1] = ub[1]/10;

			Hist2D h = new Hist2D(lb,ub,bins);

			// This is to prevent NaN's from appearing in the table.
			if (weight == 0)
				{ java.util.Arrays.fill(tab[i],0.0);
			continue;
				}
			
			for(j=-radius;j<=radius;j++)
			{
				// af.translateY(sr,j);
				sr.setShift(j);
				if (sm != null) h.set(s1,sr,sm); else h.set(s1,sr);

				double t = h.mutualInfo();

				tab[i][j+radius] = t*weight;
			}
		}

		return tab;
	}

	private int[] solveTable(double[][] table)
	{
		// These adjacency tables correspond to monotone increasing and
		// monotone decreasing sequences, respectively.

		int[][] off = {{0,1},{0,-1}};
		int[] best = null; double max=Double.NEGATIVE_INFINITY;

		for(int k=0;k<off.length;k++)
		{
			int[] out = DPSolve.solveMaxMonotonePath(table, off[k]);

			// Compute objective function to see which direction is better
			double tot=0; for(int i=0;i<out.length;i++) tot+=table[i][out[i]];
			if (tot > max) { max = tot; best = out; }
		}

		return best;
	}

	private boolean okayToFix(int[] best)
	{
		if (fix_level <= 2)
		{
			// Total number of hits must be at least max-3.
			if (hits < best.length - skippedSlices - 2) return false;
		}

		if (fix_level <= 1)
		{
			// Total absolute discrepancy must be at most 10.
			int t=0; for(int i=0; i<best.length; i++)
               if (maskArea[i]>minArea) t += Math.abs(best[i]);
			if (t > 10) return false;

			// Total change in discrepancy must at most 3.
			if (Math.abs(best[0] - best[best.length-1]) > 3) return false;

			// Average waste must be small.
			// if (waste > 20*(best.length-hits) ) return false;
		}

		return true;
	}


	public void run()
	{
		loadData();
		double[][] table = compileTable(radius);

		int[] best = solveTable(table);

		//System.out.println(Matrix.toString(table));

		for(int i=0;i<best.length;i++)
        { if (maskArea[i]>minArea) System.out.print(best[i]-radius+" ");
          else System.out.print("* ");
        }

		System.out.println();

		// Now apply heuristics to evaluate plausibility
		ArgMax am = new ArgMax();

		for(int i=0;i<best.length;i++)
		{
            if (maskArea[i] <= minArea) {skippedSlices++; continue;}

			am.argmax(table[i]);
			if (best[i] == am.arg) hits++;
			waste += am.max - table[i][best[i]];
			miss += Math.abs(am.arg - best[i]);
            total++;
		}

		System.out.println(hits + "/" + total + " hits, " + (int)waste + " waste, " + miss + " pixels out.");

		for(int i=0;i<best.length;i++) best[i]-=radius;

		if (fix_level <= 0) return;

		mn.recode('L'); mn.append("T");

		System.out.println("Fixing " + mn.getName() + "...");

        {
          int first,last;
          for(first=0;first<best.length;first++)
            if (maskArea[first] > minArea) break;
          for(last=best.length-1;last>=0;last--)
            if (maskArea[last] > minArea) break;

          if (last < 0) java.util.Arrays.fill(best,0);
          else
          {
            for(int i=0;i<first;i++) best[i]=best[first];
            for(int i=last+1;i<best.length;i++) best[i]=best[last];
          }
        }

		if (okayToFix(best))
		{
			System.out.println("Truncating to:");
		    for(int i=0;i<best.length;i++) System.out.print(best[i]+" ");
			System.out.println();

			MCVolume mv = mr.getMCVolume();
			Volume v = mv.getVolume(0);
			
			for(int i=0; i<v.getNSlices(); i++)
			{
				ShiftSlice s = new ShiftSlice(v.getSlice(i));
				s.setShift(best[i]); v.setSlice(s,i);
			}

			mv.setVolume(v,0);
			mr.setMCVolume(mv);

			mr.write(mn.toString());
		}
		else
		{
			System.out.println("Fix level " + fix_level + " exceeded, skipping.");
		}
	}
}

// Simple class to allow two return values.

class ArgMax
{
	public int arg; public double max; ArgMax() {}

	public void argmax(double[] x)
	{
		arg = 0; max = x[0];
		for(int i=1;i<x.length;i++) if (x[i] > max) { max = x[i]; arg = i; }
	}
}
