package apps.resample;

import java.util.StringTokenizer;
import java.io.*;
import cj.fileformats.MIF4;
import cj.utilities.*;
import cj.volume.*;
import cj.volume.process.Resample;
import cj.math.interpolate.*;
import cj.math.transform.*;
import cj.appkit.*;

final public class ResampleApp {
   public static void main(String argv[]) { new ResampleKit().main(argv); }
}

final class ResampleKit extends FilterAppKit 
{
	private String output_suffix = "R";
	private String comment = "";
	private VolumeResampler resampler = null;
	private double rx = 0.0;
	private double ry = 0.0;
	private double rz = 0.0;
	private double tx = 0.0;
	private double ty = 0.0;
	private double tz = 0.0;

	// Output voxel sizes
	private double[] os = {-1.0, -1.0, -1.0};

	// Output dimensions
	private int odx = -1;
	private int ody = -1;
	private int odz = -1;

	private String xInterp = "sinc";
	private String yInterp = "sinc";
	private String zInterp = "linear";

	public String getAppName() { return "resample"; }
	public String getDialogName() { return "Process image"; }
    public String getOutputSuffix() { return output_suffix; }
    public String getComment() { return comment; }

    protected void moreOpts() { 
        addOpt_reqArg("output-suffix", 800, "Output suffix (def: 'R').");

        addOpt_reqArg("rx", 801, "Rotation around X (def: 0 degrees).");
        addOpt_reqArg("ry", 802, "Rotation around Y (def: 0 degrees).");
        addOpt_reqArg("rz", 803, "Rotation around Z (def: 0 degrees).");

        addOpt_reqArg("tx", 804, "Translation around X (def: 0).");
        addOpt_reqArg("ty", 805, "Translation around Y (def: 0).");
        addOpt_reqArg("tz", 806, "Translation around Z (def: 0).");

        addOpt_reqArg("ix", 807, "Interpolator in X direction (def: sinc).");
        addOpt_reqArg("iy", 808, "Interpolator in Y direction (def: sinc).");
        addOpt_reqArg("iz", 809, "Interpolator in Z direction (def: linear).");

        addOpt_reqArg("osx", 810, "Output voxel size in X direction (def: ).");
        addOpt_reqArg("osy", 811, "Output voxel size in Y direction (def: ).");
        addOpt_reqArg("osz", 812, "Output voxel size in Z direction (def: ).");

        addOpt_reqArg("odx", 813, "Output dimension in X direction (def: ).");
        addOpt_reqArg("ody", 814, "Output dimension in Y direction (def: ).");
        addOpt_reqArg("odz", 815, "Output dimension in Z direction (def: ).");

        addOpt_reqArg("odz", 815, "Output dimension in Z direction (def: ).");

        addOpt_reqArg("like", 816, "Output same voxel size and dimension as <arg>");

        addOpt_reqArg("param", 'p', "The parameter file.");
        addOpt_reqArg("rdf-file", 'r', "The RDF file.");
    }

    protected boolean parseArg(int c)
    {
    	switch(c) {
			case 801:
				rx = getDoubleArg("rotation X");
			break;
			case 802:
				ry = getDoubleArg("rotation Y");
			break;
			case 803:
				rz = getDoubleArg("rotation Z");
			break;

			case 804:
				tx = getDoubleArg("translation X");
			break;
			case 805:
				ty = getDoubleArg("translation Y");
			break;
			case 806:
				tz = getDoubleArg("translation Z");
			break;
			
			case 807:
				xInterp = getStringArg("interpolator X");
			break;
			case 808:
				yInterp = getStringArg("interpolator Y");
			break;
			case 809:
				zInterp = getStringArg("interpolator Z");
			break;

			case 810:
				os[0] = getDoubleArg("Output Size X");
			break;
			case 811:
				os[1] = getDoubleArg("Output Size Y");
			break;
			case 812:
				os[2] = getDoubleArg("Output Size Z");
			break;

			case 813:
				odx = getIntArg("Output Dimension X");
			break;
			case 814:
				ody = getIntArg("Output Dimension Y");
			break;
			case 815:
				odz = getIntArg("Output Dimension Z");
			break;

			case 816:
				MIF4 m1 = new MIF4(getStringArg("likewise MIF"));
				os[0] = (double)m1.GetPixelX()/1000.0;
				os[1] = (double)m1.GetPixelY()/1000.0;
				os[2] = (double)m1.GetSliceThick()/1000.0;
				odx = m1.getMCVolume().getNCols();
				ody = m1.getMCVolume().getNRows();
				odz = m1.getMCVolume().getNSlices();
			break;

			case 800:
				output_suffix = getStringArg("suffix");
			break;
			case 'r':
				readRDF(getStringArg("RDF file"));
			break;
			case 'p':
				readParams(getStringArg("param file"));
			break;
			default:
				return false;
		}
		return true;

    }

	/**
	 *  Fill this one in!!
	 */
	final public void pre()
	{
		MIFName mn = new MIFName(otherArgs[0]);
        if (!mn.isValid())
          { System.out.println("Invalid MIF name "+otherArgs[0]); System.exit(-1);}

		String outfile = mn.append(output_suffix).getFile().toString();

		// Load up the data
		MIF4 m1 = new MIF4(otherArgs[0]);

		// now call register to do the registration.
		double[] vx1 = new double[3];

		vx1[0] = (double)m1.GetPixelX()/1000.0;
		vx1[1] = (double)m1.GetPixelY()/1000.0;
		vx1[2] = (double)m1.GetSliceThick()/1000.0;

		resampler = new VolumeResampler(m1.getMCVolume().getVolume(0), vx1);
		
		// If the values are not set on the command line, then
		// set it to the input volume dimensions
		if( odx < 0 ) odx = m1.getMCVolume().getNCols();
		if( ody < 0 ) ody = m1.getMCVolume().getNRows();
		if( odz < 0 ) odz = m1.getMCVolume().getNSlices();

		// If the values are not set on the command line, then
		// set it to the input volume voxel sizes
		if( os[0] < 0 ) os[0] = vx1[0];
		if( os[1] < 0 ) os[1] = vx1[1];
		if( os[2] < 0 ) os[2] = vx1[2];

		resampler.setDimensions( ody, odx, odz );
		resampler.setOutputVoxelSize( os );

		resampler.setRotationX( rx );
		resampler.setRotationY( ry );
		resampler.setRotationZ( rz );

		resampler.setTranslationX( tx );
		resampler.setTranslationY( ty );
		resampler.setTranslationZ( tz );

		try {
		resampler.setColInterpolator( InterpolateRow.createInterpolator(xInterp) );
		resampler.setRowInterpolator( InterpolateRow.createInterpolator(yInterp) );
		resampler.setSliceInterpolator( InterpolateRow.createInterpolator(zInterp) );
		} catch( UnsupportedOperationException e) { System.out.println(e); System.exit(-1);}

		Volume out = resampler.getVolume();
		System.out.println(os[0]+" "+os[1]+" "+os[2]+" "+odx+" "+ody+" "+odz);

		MCVolume v = new MCVolume(out.getNRows(),out.getNCols(),out.getNSlices(),1);
		v.setVolume(out,0);
		m1.setMCVolume(v);

        m1.SetPixelX((int)(1000.0*os[0]));
        m1.SetPixelY((int)(1000.0*os[1]));
        m1.SetSliceThick((int)(1000.0*os[2]));
		m1.write(outfile);
	}

	final public void initParams(MCVolume v) { }

    final public void printArgs() { }

    final public void run() { pre(); }

    final public MCVolume filter(MCVolume v)
		{ return v; }

	public void readParams(String param)
	{
		double[] r3t3 = new double[6];
		int[] baseDim = new int[3];
		double[] baseVox = new double[3];
		int[] dynDim = new int[3];
		double[] dynVox = new double[3];

		try{
			BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(param)));

			String str; StringTokenizer st;

			do { str = in.readLine(); }
			while (!str.startsWith("Final parameters"));

			str = in.readLine(); st = new StringTokenizer(str," ");
			for(int i=0;i<6;i++) r3t3[i] = Double.parseDouble(st.nextToken());

		} catch(Exception e)
		{
			System.out.println("Can't read params: "+ e);
			System.exit(-1);
		}

		System.out.print("Got parameters:");
		for(int i=0;i<6;i++) System.out.print(" "+r3t3[i]); System.out.println();

		rx = r3t3[0]; ry = r3t3[1]; rz = r3t3[2];
		tx = r3t3[3]; ty = r3t3[4]; tz = r3t3[5];
	}
	/**
	 *  Constructor that takes two data volumes.
	 */
	public void readRDF(String rdf)
	{
		double[] r3t3 = new double[6];
		int[] baseDim = new int[3];
		double[] baseVox = new double[3];
		int[] dynDim = new int[3];
		double[] dynVox = new double[3];

		try{
			BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(rdf)));

			String str; StringTokenizer st;

			do { str = in.readLine(); }
			while (!str.startsWith("Dim (x y z)"));

			str = in.readLine(); st = new StringTokenizer(str," ");
			for(int i=0;i<3;i++) baseDim[i] = Integer.parseInt(st.nextToken());

			do { str = in.readLine(); }
			while (!str.startsWith("Scale (x y zthick"));

			str = in.readLine(); st = new StringTokenizer(str," ");
			for(int i=0;i<3;i++) baseVox[i] = Double.parseDouble(st.nextToken());

			do { str = in.readLine(); }
			while (!str.startsWith("Dim (x y z)"));

			str = in.readLine(); st = new StringTokenizer(str," ");
			for(int i=0;i<3;i++) dynDim[i] = Integer.parseInt(st.nextToken());

			do { str = in.readLine(); }
			while (!str.startsWith("Scale (x y zthick"));

			str = in.readLine(); st = new StringTokenizer(str," ");
			for(int i=0;i<3;i++) dynVox[i] = Double.parseDouble(st.nextToken());

			do { str = in.readLine(); }
			while (!str.startsWith("Registering Parameters"));

			str = in.readLine(); st = new StringTokenizer(str," ");
			for(int i=0;i<6;i++) r3t3[i] = Double.parseDouble(st.nextToken());

		} catch(Exception e)
		{
			System.out.println("Can't read RDF: "+ e);
			System.exit(-1);
		}

		System.out.print("Got parameters:");
		for(int i=0;i<6;i++) System.out.print(" "+r3t3[i]); System.out.println();

		// MAJOR HACK TO CONVERT RDF PARAMETERS TO PROPER CENTRIC SPACE -- ERICK

		TransformAffine T = new TransformAffine();

        T.rotateZ(-r3t3[2]).rotateY(-r3t3[1]).rotateX(-r3t3[0]);
		
        T.translateX(dynVox[0]*(dynDim[0]-1)/2.0);
        T.translateY(dynVox[1]*(dynDim[1]-1)/2.0);
        T.translateZ(dynVox[2]*(dynDim[2]-1)/2.0);

        T.translateX(-dynVox[0]*(dynDim[0])/2.0);
        T.translateY(-dynVox[1]*(dynDim[1])/2.0);
        T.translateZ(-dynVox[2]*(dynDim[2])/2.0);

        T.rotateX(r3t3[0]).rotateY(r3t3[1]).rotateZ(r3t3[2]);

        T.translateX(r3t3[3]).translateY(r3t3[4]).translateZ(r3t3[5]);

        T.translateX(dynVox[0]*(dynDim[0])/2.0);
        T.translateY(dynVox[1]*(dynDim[1])/2.0);
        T.translateZ(dynVox[2]*(dynDim[2])/2.0);

        T.translateX(-baseVox[0]*(baseDim[0]-1)/2.0);
        T.translateY(-baseVox[1]*(baseDim[1]-1)/2.0);
        T.translateZ(-baseVox[2]*(baseDim[2]-1)/2.0);

		double[] x={0,0,0}, y={0,0,0};
		T.transform(x,y);

		r3t3[3] = y[0]; r3t3[4] = y[1]; r3t3[5] = y[2];

		System.out.println("Correcting to:");
		for(int i=0;i<6;i++) System.out.print(" "+r3t3[i]); System.out.println();

		rx = r3t3[0]; ry = r3t3[1]; rz = r3t3[2];
		tx = r3t3[3]; ty = r3t3[4]; tz = r3t3[5];

		os[0] = baseVox[0]; os[1] = baseVox[1]; os[2] = baseVox[2];
		odx = baseDim[0]; ody = baseDim[1]; odz = baseDim[2];
	}

}
