package apps.roivolume;

import cj.appkit.AppKit;
import cj.roi.*;
import cj.utilities.Format;

/** 
 *  This class will read in a region and print out the
 *  appropriate information.  Note, if you pass in an 
 *  LES file or ROI 0.0.1 file then you must also pass
 *  in information to do with the size of volume and the
 *  size of the voxel.
 *
 */
final public class ROIVolumeApp {
   public static void main(String argv[]) { new ROIVolumeKit().main(argv); }
}

final class ROIVolumeKit extends AppKit 
{

	/** Information about the voxels. */
	VoxelInfo vi = new VoxelInfo();

	/** Do we want to show the number of pixels? */
	private boolean show_num_pixels = false;

	/** Do we want to show the areas? */
	private boolean show_area = false;

	/** Do we want to show the volume? */
	private boolean show_volume = true;

	/** Do we want to show the full volume? */
	private boolean show_full = true;

	/** Application name */
	public String getAppName() { return "roivolume"; }

	public String getDialogName() { return "Process image"; }

	/**
	 *  Display the help.
	 */
    public void help()
	{
		System.out.println(getAppName() + " [options] <ROI filename> [<ROI filename> ...]");
		printOptions();
	}

	/**
	 *  Add options.
	 */
    protected void moreOpts() 
	{ 
		addOpt_noArg("volume",800,"Show the volume of the ROIs (def: true).");
		addOpt_noArg("area",'a',"Show the area of the ROIs (def: false).");
		addOpt_noArg("pixels",'p',"Show the number of pixels in the ROIs (def: false).");
		addOpt_reqArg("voxel-info",801,"args: xDimension yDimension zDimension xSize ySize zSize");
    }

    protected boolean parseArg(int c)
    {
    	switch(c) 
		{
			case 'a':
				show_area = true;
				break;
			case 800:
				show_volume = false;
				break;
			case 801:
				vi.xDimension = getIntArg("x dimension");
				vi.yDimension = getIntArg("y dimension");
				vi.zDimension = getIntArg("z dimension");

				vi.xSize = getDoubleArg("x size");
				vi.ySize = getDoubleArg("y size");
				vi.zSize = getDoubleArg("z size");
				break;
			case 'p':
				show_num_pixels = true;
				break;
			default:
				return false;
		}
		return true;
    }

	final public void initParams() { }
    final public void printArgs() { }

	public void run() 
	{
		VoxelInfo vi = new VoxelInfo();

		if( verbose == 0 )
		{
			System.out.print("<filename>");

			if( show_num_pixels ) System.out.print("\t<number pixels>");
			if( show_area ) System.out.print("\t<area>");
			if( show_volume ) System.out.print("\t<fracvol>");
			if( show_full ) System.out.print("\t<fullvol>");

			System.out.println("");
		}

		// Run through each ROI file.
		for(int ii=0; ii<otherArgs.length; ii++) 
		{

			/*
			 *  This is a little round-about... If we read in an ROI file
			 *  then it doesn't matter what we pass in as the VoxelInfo
			 *  as it ignores it.  If we read in an LES file, then 
			 *  it uses the VoxelInfo from the most previous ROI file.
			 *  (Though as it is implemented it is from the most recent
			 *  file read in, but as it turns out that is the most 
			 *  recent ROI file anyway... cool, eh?
			 */

			String filename = otherArgs[ii];
			RegionList b = RegionList.readRegionList(filename, vi);
			vi = b.getVoxelInfo();
//			ROIList b = (ROIList)RegionList.readRegionList(filename);

			if (verbose > 0) 
			{ 	
				System.out.println("Filename is " + filename);

				System.out.print("Total num pixels is " );
				Format.print(System.out, "%4d\n", b.getTotalPixels());

				System.out.println("Area ");
				Format.print(System.out, "%4.2f\n ", b.getTotalArea());

				System.out.println("Volume ");
				Format.print(System.out, "%4.2f\n ", b.getTotalVolume());
			}
			else
			{
				System.out.print(filename);
				if( show_num_pixels ) Format.print(System.out, "\t%4d",(int)b.getTotalPixels());
				if( show_area ) Format.print(System.out, "\t%4.2f", b.getTotalArea());
				if( show_volume ) Format.print(System.out, "\t%4.2f", b.getTotalVolume());
				if( show_full ) Format.print(System.out, "\t%4.2f", b.getTotalPixels()*b.getVoxelVolume());
				System.out.println("");
			}
		}
	}
}
