package apps.t2simulate;

import java.util.Vector;
import cj.mri.relax.simulate_noisey.T1T2Simulate;
import cj.mri.relax.t2fit.*;
import cj.mri.relax.t2bin.*;
import cj.utilities.Format;
import cj.appkit.AppKit;

final public class T2SimulateApp {
	public static void main(String argv[]) 
	{ 
		String[] bob = new String[ argv.length + 1];
		for(int ii=0; ii<argv.length; ii++) bob[ii] = new String(argv[ii]);
		bob[argv.length] = "adsf";
		new T2SimulateKit().main(bob); 
	}
}


class T2SimulateKit extends AppKit
{
	/** Number of iterations over the noise. */
	private int niters = 100;

	/** The T2 bin information. */
	private Vector t2bins = new Vector();

	/** The T2 bin information. */
	private double true_alpha = 180.0;

	/** The T2 Water Compartments. */
	private Vector t2compartments = new Vector();

	/** Do we want to estimate the flip angle? */
	private boolean estimateFlip = false;

	/** The signal-to-noise ratio. */
	private double snr = 200.0;

	/** Type of noise to add. */
	private String noise_type = "rice";

	/** Random number seed. */
	private int seed = -1;

	/** The TE times. */
	private double te = 10.0;

	/** The TR times. */
	private double tr = Double.POSITIVE_INFINITY;

	/** The number of echoes. */
	private int nechoes = 32;

	/** Method from which to estimate the flip angle. */
	private String estimateMethod = "golden";

	/** Method from which to fit the decay curve. */
	private Vector fitMethod = null;
	private Vector fitMethod_names = null;

	private T1T2Simulate tts = null;

	private double biexp_first = 0.0, biexp_second = 0.0;
	private double nlower = 0.0, nupper = 0.0;

	public void help() 
	{ 
		System.out.println(getAppName() + " [Options]");
		printOptions();
		
		System.out.println("Examples:");
		System.out.println("WM: -i 100 -b s 0 50 -b m 50 120 -c 200 1000 20 -c 800 1000 80 -f nnls ");
	}

	public String getDialogName() { return ""; }
	public String getAppName() { return "t2simulate"; }

	protected void moreOpts()
	{
		addOpt_reqArg("fit-method", 'f', "Decay curve fit method (def: nnls, must be one of 'nnls', 'nnls_small <arg> <arg>', 'l1', 'l1_small <arg> <arg>','monoexp', 'monoexp_nonlin', 'monoexp_nnls', 'biexp', 'biexp_fixed <arg> <arg>','biexp_sweep')");
		addOpt_reqArg("niters", 'i', "Number of noise iterations (def: 100)");
		addOpt_reqArg("bin", 'b', 3, "Define a T2 bin (def: none)");
		addOpt_reqArg("alpha", 'a', "Set the true refocusing pulse flip angle (def: 180)");
		addOpt_reqArg("compartment", 'c', 3, "Add a T2 water compartment (def: none)");
		addOpt_noArg("estimate-flip", 'e', "Estimate the flip angle (def: false)");
		addOpt_reqArg("snr", 's', "SNR (def: 200)");
		addOpt_reqArg("noise-type", 801, "Type of noise to add ('gaussian' or 'rice', def: 'rice')");
		addOpt_reqArg("te", 't', "TE (def: 10.0 ms)");
		addOpt_reqArg("tr", 'r', "TR (def: infinity)");
		addOpt_reqArg("nechoes", 'n', "Number of echoes (def: 32)");
		addOpt_reqArg("estimate-flip-method", 800, "Flip estimation method (def: minimize, must be one of 'minimize', 'spline', 'quadratic', 'golden')");
		addOpt_reqArg("random-seed", 802, "Set the seed for the random number generator.");
	}

	protected boolean parseArg(int c)
	{
		if( t2compartments == null )
		{
			t2compartments = new Vector();
		}

		if( t2bins == null )
		{
			t2bins = new Vector();
		}

		if( tts == null ) 
		{
			tts = new T1T2Simulate();
		}

		switch (c)
		{
			case 'i': 
				niters = getIntArg("number iterations");
				break;
			case 's': 
				snr = getDoubleArg("SNR");
				break;
			case 't': 
				te = getDoubleArg("TE");
				break;
			case 'n': 
				nechoes = getIntArg("number of echoes");
				break;
			case 800: 
				estimateMethod = getStringArg("flip estimation method");
				break;
			case 801: 
				noise_type = getStringArg("type of noise");
				break;
			case 802: 
				seed = getIntArg("random number generator seed");
				break;
			case 'e': 
				estimateFlip = true;
				break;
			case 'f': 
				addFitMethod( getStringArg("decay curve fit method") ) ;
				break;
			case 'a': 
				true_alpha = getDoubleArg("true alpha");
				break;
			case 'b':
				String name = getStringArg("bin name");
				t2bins.add( new T2Bin( getDoubleArg("bin start"), 
									   getDoubleArg("bin stop"),
									   name)); 
				break;
			case 'c':
				WaterCompartment wc = new WaterCompartment();
				wc.pd = getDoubleArg("proton density");
				wc.t1 = getDoubleArg("T1");
				wc.t2 = getDoubleArg("T2");
				t2compartments.add( wc );
				break;
		}

		return true;
	}

	public void run()
	{
		long start, end;

		// Check to make sure they defined at least one method.
		if( fitMethod == null || fitMethod.size() == 0 )
		{
			System.out.println("t2simulate: You must define at least one fit algorithm");
			System.exit(0);
		}

		T2Distribution distribution = new T2Distribution();

		// Set some parameters
		tts.setNoiseType(noise_type);
		tts.setAlpha(true_alpha);
		tts.setSNR(snr);
		tts.setTE(te);
		tts.setNumEchoes(nechoes);

		// If we want to see the seed, then set the seed.
		if( seed > 0 )
		{
			tts.setRandomSeed(seed);
		}

		/*
		 *  Add the T2 water compartments specified.
		 */
		for(int ii=0; ii<t2compartments.size(); ii++)
		{
			tts.addComponent( 
				((WaterCompartment)t2compartments.get(ii)).pd,
				((WaterCompartment)t2compartments.get(ii)).t1,
				((WaterCompartment)t2compartments.get(ii)).t2 );
		}

		double[] decay = new double[ nechoes ];

		// Set the flip angle estimation method.
		for(int fi=0; fi<fitMethod.size(); fi++) 
		{
			T2Fit fitter = (T2Fit)fitMethod.get(fi);

			fitter.setEstimateFlip(estimateFlip);
			if( estimateFlip )
			{
				if( estimateMethod.compareTo("spline") == 0 )
				{
					fitter.setEstimateFlipMethod( T2Fit.ESTIMATE_FLIP_SPLINE );
				}
				else if( estimateMethod.compareTo("quadratic") == 0 )
				{
					fitter.setEstimateFlipMethod( T2Fit.ESTIMATE_FLIP_QUADRATIC );
				}
				else if( estimateMethod.compareTo("minimize") == 0 )
				{
					fitter.setEstimateFlipMethod( T2Fit.ESTIMATE_FLIP_MINIMIZE );
				}
				else if( estimateMethod.compareTo("golden") == 0 )
				{
					fitter.setEstimateFlipMethod( T2Fit.ESTIMATE_FLIP_GOLDEN );
				}
				else
				{
					System.out.print("The flip angle estimation method you ");
					System.out.println("specified " + estimateMethod + " is not known.");
				}
			}
		}

		// Now output the information.
		if( verbose > 0 )
		{
			System.out.println("Parameters were:");
			System.out.println("\tNumber of iterations: " + niters);
			System.out.println("\tTrue flip angle: " + true_alpha);
			if( estimateFlip ) 
			{
				System.out.println("\tEstimate the flip angle (method:" + estimateMethod + ")");
			}
			else System.out.println("\tNo flip angle estimation");
			System.out.println("\tSignal to noise: " + snr);
			System.out.println("\tEcho time: " + te);
			System.out.println("\tNumber echoes: " + nechoes);
			System.out.println("\tCurve fit method: " + fitMethod);

			System.out.println("Water Compartments");
			for(int ii=0; ii<t2compartments.size(); ii++)
			{
				System.out.print("\tCompartment " + (ii+1) + ":");
				System.out.print(" pd = " + ((WaterCompartment)t2compartments.get(ii)).pd);
				System.out.print(" t1 = " + ((WaterCompartment)t2compartments.get(ii)).t1);
				System.out.print(" t2 = " + ((WaterCompartment)t2compartments.get(ii)).t2);
				System.out.println("");
			}

			System.out.println("Bins");
			for(int bin=0; bin<t2bins.size(); bin++)
			{
				System.out.print("\tBin " + (bin+1) + " ");
				System.out.print( ((T2Bin)t2bins.get(bin)).getName());
				System.out.print( ": (" + ((T2Bin)t2bins.get(bin)).getStart());
				System.out.print( ", " + ((T2Bin)t2bins.get(bin)).getStop());
				System.out.println(")");
			}
		}

		double[][][] fr = new double[ fitMethod.size() ][ t2bins.size() ][ niters ];
		double[][][] dn = new double[ fitMethod.size() ][ t2bins.size() ][ niters ];
		double[][][] gm = new double[ fitMethod.size() ][ t2bins.size() ][ niters ];
		double[][] flip = new double[ fitMethod.size() ][ niters ];
		double[][] chi2 = new double[ fitMethod.size() ][ niters ];
		double[][] time = new double[ fitMethod.size() ][ niters ];

		long total_start = System.currentTimeMillis();
		for(int ii=0; ii<niters; ii++)
		{
			if( verbose < 2 )
			{
//				System.out.print("\rIteration: " + ii + "/" + niters);
			}

			// Create the decay 
			tts.getDecay( decay );

			double[] tes = new double[ nechoes ];
			for(int ti=0; ti<nechoes; ti++) tes[ti] = te*(double)(ti+1);

			// Do the solution for each fitter.
			for(int fi = 0; fi < fitMethod.size(); fi++) 
			{
				T2Fit fitter = (T2Fit)fitMethod.get(fi);
				double[] stds = tts.getStdDev();
				fitter.setStdDev( stds );
				fitter.setTE( tes );
				fitter.setTR( tr );
				//fitter.setStdDev( tts.getStdDev() );

				// Solve 
				start = System.currentTimeMillis();
				chi2[fi][ii] = fitter.solve(decay, distribution);
				end = System.currentTimeMillis();

				// If estimate flip then get best flip angle.
				if( estimateFlip )
				{
					flip[fi][ii] = fitter.getBestAlpha();
				}

				/* 
				 *  Calculate and display for each bins.
				 */
				for(int bin=0; bin<t2bins.size(); bin++)
				{
					((T2Bin)t2bins.get(bin)).calculate(distribution);

					fr[fi][bin][ii] = ((T2Bin)t2bins.get(bin)).getFR();
					dn[fi][bin][ii] = ((T2Bin)t2bins.get(bin)).getDN();
					gm[fi][bin][ii] = ((T2Bin)t2bins.get(bin)).getGM();

					if( verbose > 1 )
					{
						System.out.print( ((T2Bin)t2bins.get(bin)).getName() + ": ");
						Format.print(System.out, "%4.3f ", 
							((T2Bin)t2bins.get(bin)).getFR());
						Format.print(System.out, "%4.3f ", 
							((T2Bin)t2bins.get(bin)).getDN());
						Format.print(System.out, "%4.3f ", 
							((T2Bin)t2bins.get(bin)).getGM());
					}
				}

				// Show the chi2.
				if( verbose > 1 )
				{
					Format.print(System.out, "chi2: %4.3f", chi2[fi][ii]);
					System.out.print(" ");
				}

				// Show the flip angle, if calculated.
				if( estimateFlip && verbose > 1 )
				{
					Format.print(System.out, "flip: %4.3f", flip[fi][ii]);
					System.out.print(" ");
				}

				// Show the time.
				time[fi][ii] = (end-start)/1000.0;
				if( verbose > 1 )
				{
					Format.print(System.out, "time: %4.3f", time[fi][ii]);
					System.out.println(" ");
				}
			}
		}
		long total_end = System.currentTimeMillis();

		if( verbose < 2 )
		{
//			System.out.print("\r                     \r");
		}

		// Now do the summary report.
		System.out.println("");


		double totalpd = 0.0;
		for(int ci=0; ci<t2compartments.size(); ci++) 
			totalpd += ((WaterCompartment)t2compartments.get(ci)).pd;

		// For each fitting method compute all the stats.
		for(int fi=0; fi < fitMethod.size(); fi++)
		{
			System.out.println("Fit method: " + (String)fitMethod_names.get(fi) );

			for(int bin=0; bin<t2bins.size(); bin++)
			{

				//  Need to calculate the true pd and t2 for the bin based
				//  on the compartments that were passed in.
				double total_water = 0.0;
				double true_water = 0.0;
				double true_t2 = 0.0;
				for(int ci=0; ci<t2compartments.size(); ci++)
				{
					total_water += ((WaterCompartment)t2compartments.get(ci)).pd;
					if( ((WaterCompartment)t2compartments.get(ci)).t2 >  
					    ((T2Bin)t2bins.get(bin)).getStart() && 
						((WaterCompartment)t2compartments.get(ci)).t2 <=  
						((T2Bin)t2bins.get(bin)).getStop() )
					{
						true_water += ((WaterCompartment)t2compartments.get(ci)).pd;
						true_t2 += ((WaterCompartment)t2compartments.get(ci)).pd * ((WaterCompartment)t2compartments.get(ci)).t2;
					}

				}
				true_t2 /= true_water;

				System.out.print( ((T2Bin)t2bins.get(bin)).getName() + " -> ");
				Format.print(System.out, "fr: %4.3f ", calcMean( fr[fi][bin] ) );
				Format.print(System.out, "(%4.3f) ", calcStd( fr[fi][bin] ) );
				Format.print(System.out, "[%4.1f]  ", calcRe( fr[fi][bin], true_water/totalpd) );
					

				Format.print(System.out, "dn: %4.3f ", calcMean( dn[fi][bin] ) );
				Format.print(System.out, "(%4.3f) ", calcStd( dn[fi][bin] ) );
				Format.print(System.out, "[%4.1f]  ", 
					calcRe( dn[fi][bin],  true_water) );

				Format.print(System.out, "gm: %4.3f ", calcMean( gm[fi][bin] ) );
				Format.print(System.out, "(%4.3f) ", calcStd( gm[fi][bin] ) );
				Format.print(System.out, "[%4.1f]\n", calcRe( gm[fi][bin],  true_t2) );
			}
			System.out.println("");
		
			if( estimateFlip )
			{
				Format.print(System.out, "Flip: %4.3f", calcMean( flip[fi] ) );
				Format.print(System.out, " (%4.3f)", calcStd( flip[fi] ) );
				Format.print(System.out, " [%4.3f]", calcRe( flip[fi], true_alpha ));
				System.out.println("");
			}

			Format.print(System.out, "Misfit: %4.3f", calcMean( chi2[fi] ) );
			Format.print(System.out, " (%4.3f)", calcStd( chi2[fi] ) );
			Format.print(System.out, " (normalized misfit %4.3f)", calcMean( chi2[fi] ) / nechoes );
			System.out.println("");
			Format.print(System.out, "Time per solution (s): %4.3f", calcMean( time[fi] ) );
			Format.print(System.out, " (%4.3f)", calcStd( time[fi] ) );
			System.out.println("");
			Format.print(System.out, "Total time (s): %4.3f", 
				(double)(total_end-total_start)/1000.0);
			System.out.println("");
			System.out.println("");
		}
	}

	/**
	 *  Compute the mean of any vector.
	 */
    private final double calcMean(final double[] data)
    {
        double tt = 0.0;
        for(int ii=0; ii<data.length; ii++)
        {
            tt += data[ii];
        }

        return tt/(double)data.length;
    }

	/**
	 *  Compute the relative error of any vector.
	 */
    private final double calcRe(final double[] data, double true_value)
    {
        double tt = 0.0;
        for(int ii=0; ii<data.length; ii++)
        {
            tt += (data[ii]-true_value)/true_value*100.0;
        }

        return tt/(double)data.length;
    }

	/**
	 *  Compute the standard deviation of any vector.
	 */
    private final double calcStd(final double[] data)
    {
        double tt = 0.0;
        double tt2 = 0.0;
        int N = data.length;
        for(int ii=0; ii<N; ii++)
        {
            tt += data[ii];
            tt2 += data[ii]*data[ii];
        }

        return Math.sqrt((tt2-(tt*tt)/(double)N)/(double)(N-1) );
    }

	/**
	 *  Add a fitting routine to the list of routines.
	 */
	private void addFitMethod( final String fitname )
	{
		// Create the vectors if not created yet.
		if( fitMethod == null )
		{
			fitMethod = new Vector();
			fitMethod_names = new Vector();
		}

		fitMethod_names.add( fitname );

		// Figure out what type it is.
		if( fitname.compareTo("nnls") == 0 ) 
		{
			fitMethod.add( new T2FitNNLS( tts.getTEList() ) );
		}
		else if( fitname.compareTo("nnls_small") == 0 ) 
		{
			nlower = getDoubleArg("nnls_small first parameter");
			nupper = getDoubleArg("nnls_small second parameter");
			
			T2Fit temp = new T2FitNNLS( tts.getTEList() );
			double[] tt = {nlower, nupper};

			temp.setParameters(tt);
			fitMethod.add( temp );
		}
		else if( fitname.compareTo("l1") == 0 ) 
		{
			fitMethod.add( new T2FitL1( tts.getTEList() ) );
		}
		else if( fitname.compareTo("l1_small") == 0 ) 
		{
			nlower = getDoubleArg("l1_small first parameter");
			nupper = getDoubleArg("1l_small second parameter");
			
			T2Fit temp = new T2FitL1( tts.getTEList() );
			double[] tt = {nlower, nupper};

			temp.setParameters(tt);
			fitMethod.add( temp );
		}
		else if( fitname.compareTo("monoexp") == 0 ) 
		{
			fitMethod.add( new T2FitMonoExp( tts.getTEList() ) );
		}
//		else if( fitname.compareTo("monoexp_nonlin") == 0 ) 
//		{
//			fitMethod.add( new T2FitMonoExpNonLinear( tts.getTEList() ) );
//		}
		else if( fitname.compareTo("monoexp_nnls") == 0 ) 
		{
			fitMethod.add( new T2FitMonoExpNNLS( tts.getTEList() ) );
		}
		else if( fitname.compareTo("biexp") == 0 ) 
		{
			fitMethod.add( new T2FitBiExp( tts.getTEList() ) );
		}
		else if( fitname.compareTo("biexp_sweep") == 0 ) 
		{
			fitMethod.add( new T2FitBiExpSweep( tts.getTEList() ) );
		}
		else if( fitname.compareTo("biexp_fixed") == 0 ) 
		{
			biexp_first = getDoubleArg("biexp first parameter");
			biexp_second = getDoubleArg("biexp second parameter");

			T2Fit temp = new T2FitBiExpFixedT2( tts.getTEList() );
			double[] tt = {biexp_first, biexp_second};
			temp.setParameters(tt);
			fitMethod.add( temp );
		}
		else
		{
			System.out.println("t2simulate: Unknown fit method " + fitname);
			System.exit(-1);
		}
	}
}

class WaterCompartment
{
	public double pd=0.0;
	public double t2=0.0;
	public double t1=0.0;
}
