package apps.tecorrect;

import gnu.getopt.*;
import cj.fileformats.*;
import cj.volume.MCVolume;
import cj.volume.measure.*;
import cj.volume.process.TECorrect;
import cj.appkit.FilterAppKit;

final public class TECorrectApp {
   public static void main(String argv[]) {
      TECorrectKit myApp = new TECorrectKit(); myApp.main(argv);
   }
}

final class TECorrectKit extends FilterAppKit {

	private double[] curr_tes = new double[2];
	private TECorrect tec = null;
    private String comment;
	private double[] tes = new double[20];  // Ok, ok, it's hard coded.
	private int num_tes = 0;

	public String getAppName() { return "tecorrect"; }
	public String getDialogName() { return "Correct images..."; }
    public String getOutputSuffix() { return ""; }
    public String getComment() { return comment; }

    public void help() {
		System.out.println("tecorrect [options] <filename> [<filename> ...]");
		printOptions();
    }

    protected void moreOpts() { 
		num_tes=0;

        addOpt_reqArg("te", 't', "Echo Time (ms)");
    }

    protected void parseArg(int c, Getopt g)
    {
    	switch(c) {
			case 'v':
				verbose++;
			break;
			case 't':
				tes[num_tes] = getDoubleArg("tes");
				num_tes++;
			break;
			default:
				help();
                System.exit(-1);
			break;
		}
    }

	final public void initParams(MCVolume v)
	{
		tec = new TECorrect(tes);
		tec.setVerbose(verbose);
	}

    final public void printArgs()
   	{
		System.out.print("Desired echo times are: " );
		for(int ii=0; ii<num_tes; ii++) 
		{
			System.out.print(tes[ii]);

			if( ii < num_tes-1 )
			{
				System.out.print(", ");
			}
		}

		System.out.println("");
   	}

    final public MCVolume filter(MCVolume v)
	{
		if( curr_tes.length != num_tes )
		{
			System.out.print("The number of TEs you entered on the ");
			System.out.print("command line was " + num_tes + ", but ");
			System.out.print("the number of TEs in the file is " );
			System.out.println(curr_tes.length + "! They must be the same! ");
			System.exit(-1);
		}

		tec.monoCorrect(v, curr_tes);
		return v;
	}

	public void run() 
	{
        pre();
		MIF4 ms = null;
		MIF4 ml = null;
		String long_filename = "";

		for(int ii=0; ii<otherArgs.length; ii++) 
		{
			filename = otherArgs[ii];

			if (verbose > 0) { System.out.println("Filename is " + filename); }

			ms = new MIF4(filename);
			MCVolume vols = ms.getMCVolume();
			double[] tt = ms.getTE();
			curr_tes[0] = tt[0];

			//
			//  If there is only one channel, then we likely need to read in 
			//  the other echo (ie that it is a single echo MIF file).
			//
			if( vols.getNChannels() == 1 )
			{
				long_filename = createLongFilename(filename);

				ml = new MIF4(long_filename);
				MCVolume newvol = new MCVolume(vols.getNRows(), vols.getNCols(),
					vols.getNSlices(), 
					vols.getNChannels()+ml.getMCVolume().getNChannels());

				newvol.setVolume(vols.getVolume(0), 0);
				newvol.setVolume(ml.getMCVolume().getVolume(0), 1);
				vols = newvol;

				tt = ml.getTE();
				curr_tes[1] = tt[0];
			}

			if( verbose > 0 ) 
			{
				System.out.print("Current TE's are: ");
				for(int kk=0; kk<curr_tes.length; kk++) 
				{
					System.out.print(curr_tes[kk]);
					if( kk < curr_tes.length-1 ) { System.out.print(", "); }
				}
				System.out.println("");
			}

			initParams(vols);

            if (verbose > 0) { printArgs(); }

            filter(vols);

			if (filterMode == READ_WRITE)
			{
				MCVolume temp = new MCVolume(vols.getNRows(), vols.getNCols(), vols.getNSlices(), 1);
				temp.setVolume(vols.getVolume(0), 0);
				ms.setVolume(temp);
				ms.write(filename);

				temp.setVolume(vols.getVolume(1), 0);
				ml.setVolume(temp);
				ml.write(long_filename);
			}
		}

		post();

        // Temporary hack for java not exiting due to persistent JFrame(?)
		System.exit(0);
	}

	/**  Create a long echo MIF filename based on this short
	 *   echo MIF filename.
	 */
	private String createLongFilename(String filename)
	{
		//
		//  First, let's make sure the thing is a MIF file.
		//
		if( !filename.toUpperCase().endsWith(".MIF") && 
		    !filename.toUpperCase().endsWith(".MAF") &&
		    !filename.toUpperCase().endsWith(".M2F") )
		{
			System.out.println("The file, " + filename + ", must be a MIF file.");
			System.exit(-1);
		}

		//
		//  Now let's make sure that it is a short echo file.
		//
		int pos = filename.lastIndexOf('/'); pos++;

		System.out.println("pos = " + pos);
		pos = filename.toUpperCase().indexOf('S', pos);
		if( pos <= 0 )
		{
			System.out.println("pos = " + pos + " and " + filename);
			System.out.println("The file does not seem to be a short echo MIF file.");
			System.exit(-1);
		}

		String filename_long = filename.substring(0, pos) + "L" + 
		                       filename.substring(pos+1);

		return filename_long;
	}
}
