package cj.mri.relax.t2map;

import java.util.*;
import javax.swing.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import cj.volume.*;
import cj.volume.measure.NoiseEstimate;
import cj.mri.relax.t2bin.*;
import cj.mri.relax.t2fit.*;

/**
 *  The class is the base class for all other T2Map classes
 *  and takes care of most things including storing the spectral
 *  volume, chi2 map, variable flip angle map (if needed) and estimating
 *  the noise in the image.
 *
 */
abstract public class T2Map
{
	protected int verbose = 0;

	// Noise Estimators.

	/** No noise estimation. */
	public static final int NE_NONE = 1;

	/** Estimate the noise across the rows */
	public static final int NE_ACROSS_ROW = 2;

	/** Estimate the noise across the columns. */
    public static final int NE_ACROSS_COLUMN = 3;

	/** Estimate the noise in the image (gradientMethod). */
    public static final int NE_ACROSS_IMAGE = 4;
    private int noise_estimate = NE_ACROSS_ROW;

	// Used to speed up the standard deviation calculations
	private int ne_prev_row = -1;
	private int ne_prev_col = -1;
	private boolean ne_image_calculated = false;

	/** The T2 fitting object. */
	protected T2Fit t2fitter = null;

    /** Amplitude solutions to NNLS */
    protected SpectralVolume svol = null;
 
    /** Chi2 solution to NNLS */
    protected MCVolume chi2 = null;

    /** Best estimated flip angle if estimateFlip is true. */
    protected MCVolume alpha_map = null;

	// NOTE:  estimateFlip AND useFlip CAN'T BOTH BE TRUE!

    /** Determine if we are going to estimate the flip angle. */
    protected boolean estimateFlip = false;
	
    /** Determine if we are going to estimate the flip angle. */
    protected boolean useFlip = false;
	
	/** The threshold on the signal intensity. */
	protected double threshold = 20;

	/** Myelin slice, if the user wants to display intermediate results. */
	protected MCVolume myelin_slice = null;
	protected boolean show_myelin_slice = false;
	protected JFrame frame = null;
	protected MCVolumePanel dm = null;

	/** Progress bar, if the user wants to display intermediate results. */
	protected boolean show_progress_bar = false;
	protected JFrame progressFrame = null;
	protected JProgressBar progressBar = null;

	public T2Map()
	{
	}

	/*
	 *  Abstract methods....
	 *
	 */

	abstract public void setTE( double[] te );

	/** Return the number of spectral channels to be defined for the 
	 *  spectral volume. 
	 */
	abstract protected int getNSpectralChannels();

	/** Deal with the distribution and chi2. */
	abstract protected void dealWithData(final T2Distribution dist, 
		final double chi2, int row, int col, int slice);

	/** Deal with the flip angle that was estimated. */
	abstract protected void dealWithAlpha(final double flip, 
		int row, int col, int slice);

	/** Save the calculated data to a file. */
	abstract public void saveData( String base, Vector header, final String comment );

	/** Set the parameters, if there are any... */
	abstract public void setParameters( double[] params );

	/** Allow the person to show the myelin map as we calculate it. */
	public void setInteractiveMyelin( final boolean yup )
	{
		show_myelin_slice = yup;
	}

	/** Set the threshold... */
	public void setThreshold( final double threshold )
	{
		this.threshold = threshold;
	}

	/** Set verbose */
	public void setVerbose( final int verbose )
	{
		this.verbose = verbose;	
		t2fitter.setVerbose( verbose );
	}

	/** Should we do flip angle estimation? */
	public void setEstimateFlip( final boolean estimateFlip )
	{
		if( useFlip )
		{
			System.out.println("t2map: Can't use a flip angle map *and* estimate the flip angle at the same time.");
			System.exit(-1);
		}


		this.estimateFlip = estimateFlip;
		t2fitter.setEstimateFlip( estimateFlip );
	}

	/** Should we do flip angle estimation? */
	public void setUseFlip( final MCVolume alpha_map )
	{
		if( estimateFlip )
		{
			System.out.println("t2map: Can't estimate flip angle *and* use a flip angle map at the same time.");
			System.exit(-1);
		}

		this.alpha_map = (MCVolume)alpha_map.clone();

		useFlip = true;
	}

	/**
	 *  Apply the t2fit algorithm over the whole volume.
	 */
    public void map(final MCVolume vol)
    {
        map(vol, 0, vol.getNRows()-1, 0, vol.getNCols()-1,
                        0, vol.getNSlices()-1);
    }
 
	/**
	 *  Apply the t2fit algorithm over the specified sub volume.
	 */
	public void map(final MCVolume vol, 
						final int start_row, final int end_row,
						final int start_col, final int end_col,
						final int start_slice, final int end_slice)
	{
		T2Distribution dist = new T2Distribution();
		T2Bin myelin_bin = new T2Bin(0.0, 50.0, "short");

		double[] stds = new double[ vol.getNChannels() ];
		double chi2_temp;

		svol = new SpectralVolume(vol.getNRows(), vol.getNCols(), 
									 vol.getNSlices(), getNSpectralChannels());
		chi2 = new MCVolumeFile(vol.getNRows(), vol.getNCols(), 
									 vol.getNSlices(), 1);


		/*
		 *  If we are going to show the slice, then create the frame and
		 *  MCVolumePanel to display it.
		 */
		if( show_myelin_slice )
		{
			try {
				frame = new JFrame("Myelin Image");
				frame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );

				// Set the initial location in the middle of the screen.
				frame.setSize(256+5,256+40);
				Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
				frame.setLocation(screen.width/2-256/2,screen.height/2-256/2);

				int nRows=vol.getNRows(), nCols=vol.getNCols();
				frame.setSize(nCols+5,nRows+40);
				frame.setLocation(screen.width/2-nCols/2,screen.height/2-nRows/2);

				frame.setResizable(true);
				frame.show();
			}
			catch( InternalError e )
			{ 
				System.out.println("Couldn't open a window, continuing anyway");
			  	show_myelin_slice = false;
			}
		}

		if( estimateFlip )
		{
			alpha_map = new MCVolumeFile(vol.getNRows(), vol.getNCols(), 
									 vol.getNSlices(), 1);
		}

		if( show_progress_bar)
		{
			try{
				progressFrame = new JFrame("Progress...");
				progressFrame.setSize(256,50);
				progressFrame.setResizable(false);
				progressBar = new JProgressBar(0, 100); 
				progressBar.setValue(0);
				progressBar.setMaximum(100);
				progressFrame.getContentPane().add( progressBar );
				Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
				progressFrame.setLocation(screen.width/2-128,screen.height/2-128);
				progressFrame.show();
			}
			catch( InternalError e )
			{
				System.out.println("Couldn't open a window, continuing anyway");
			  	show_progress_bar = false;
			}
		}

		for(int slice=start_slice; slice < end_slice; slice++) 
		{

			// Are we going to display while we are doing the stuff?
			if( show_myelin_slice )
			{
				myelin_slice = new MCVolume(vol.getNRows(), vol.getNCols(), 1, 1);

				dm = new MCVolumePanel(myelin_slice, false, true);
				frame.getContentPane().add(dm);
			}

			for(int row=start_row; row < end_row; row++) 
			{
				if( verbose > 1 ) 
				{
					System.out.println("-------------------------------------");
				}

				System.out.print("\rrow/slice = " + row + "/" + slice);

				if( verbose > 1 ) 
				{
					System.out.println("");
				}

				/*
				 *  Calculate the nnls solution for each row.
				 */
				for(int col=start_col; col < end_col; col++) 
				{
					if( show_progress_bar)
					{
						progressBar.setValue( 100*
							((slice-start_slice)*(end_row-start_row)*(end_col-start_col) + 
							 (row-start_row)*(end_col-start_col) + col-start_col) / 
							 ((end_row-start_row)*(end_col-start_col)*(end_slice-start_slice))
						);
					}

					//  Compute the standard deviation for each channel.
					calculateStd(vol, row, col, slice, stds);

					// Set the standard deviations.
					t2fitter.setStdDev( stds );

					// Make sure we are above the threshold.
					if( vol.getData(row,col,slice,0) > threshold*stds[0] ) 
					{

						// If we are using a flip angle map, then set
						// the flip angle.
						if( useFlip )
						{

							double alphatt = alpha_map.getData(row,col,slice,0);
							if( alphatt < 2.0 )
							{
								t2fitter.setAlpha( 180.0 );
							}
							else
							{
								t2fitter.setAlpha( alphatt );
							}
						}

						chi2_temp = t2fitter.solve(vol.getChannelDouble(row,col,slice), dist);	
//						chi2_temp = t2fitter.solveBootstrap(vol.getChannelDouble(row,col,slice), dist);	

						// Are we going to display while we are doing the stuff?
						if( show_myelin_slice )
						{
							myelin_bin.calculate(dist);

							if( myelin_bin.getFR() < 0.8 )
							myelin_slice.setData(myelin_bin.getFR(),row,col,0,0);
						}

						// Put results in proper place.
						dealWithData(dist, chi2_temp, row,col,slice);

						// If we must do something with the estimated 
						// flip angle, then do it.
						if( estimateFlip )
						{
							dealWithAlpha(t2fitter.getBestAlpha(),
                                  row,col,slice);
						}

						// Show what the result is if we really, really 
						// want to.
						if( verbose > 1 )
						{
							double[] x = dist.getAmplitudes();
							double[] t2s = dist.getT2s();
							for(int ii=0; ii<x.length; ii++) 
							{
								if( x[ii] > 0.0 )
								{
									System.out.println(x[ii] + "  " + t2s[ii]);
								}
							}
							System.out.println("chi2 = " + chi2_temp);
						}
					}
				}

				// Are we going to display while we are doing the stuff?
				if( show_myelin_slice )
				{
					// This is the old way and might be a little too slow.
//					frame.getContentPane().remove(dm);
//					dm = new MCVolumePanel(myelin_slice, false, true);
//					frame.getContentPane().add(dm);

					dm.setRow(myelin_slice.getRow(row,0,0),row,0,0);
				}

			}
		}
	}

	/** 
	 *  Compute the standard deviation in the specified way.
	 */
	private boolean calculateStd(final MCVolume vol, 
		final int row, final int col, final int slice, double[] stds)
	{
		boolean updated = false;

		switch( noise_estimate )
		{
			case T2Map.NE_NONE:
				Arrays.fill(stds, 1.0);
				break;
			case T2Map.NE_ACROSS_ROW:
				if( ne_prev_row != row )
				{
					stdAcrossRow(vol, row, slice, stds);
					ne_prev_row = row;
					updated = true;
				}
				break;
			case T2Map.NE_ACROSS_COLUMN:
				if( ne_prev_col != col )
				{
					stdAcrossColumn(vol, col, slice, stds);
					ne_prev_col = col;
					updated = true;
				}
				break;
			case T2Map.NE_ACROSS_IMAGE:
					if( !ne_image_calculated )
					{
						stdAcrossImage(vol, slice, stds);
						ne_image_calculated = true;
					}
				break;
			default:
				System.out.println("T2Map: Unknown noise estimate type.");
				System.exit(-1);
				break;
		}

		return updated;
	}

	/** stdRowSlice - Compute the standard deviation across a row
	 *                given the row and slice.  Do this for each
	 *                channel.
	 */
	protected void stdAcrossRow(final MCVolume vol, final int row, 
								 final int slice, double[] stds)
	{
		int n = 0;
		double x = 0.0;
		double x2 = 0.0;
		double datum = 0;

		for(int ch=0; ch < vol.getNChannels(); ch++) 
		{
			n = 0;
			x = x2 = 0.0;
			for(int col=0; col < vol.getNCols(); col++) 
			{
				datum = vol.getData(row,col,slice,ch);

				if( datum < 50.0 )
				{
					x += datum;
					x2 += (datum*datum);
					n++;
				}
			}

			if( x == 0 || n == 0 )
			{
				stds[ch] = 1.0;
			}
			else
			{
				stds[ch] = Math.sqrt((double)(x2 - (x*x)/(double)n)/(double)(n-1));
			}
		}
	}

	/** stdColSlice - Compute the standard deviation across a column
	 *                given the column and slice.  Do this for each
	 *                channel.
	 */
	protected void stdAcrossColumn(final MCVolume vol, final int col, 
								 final int slice, double[] stds)
	{
		int n = 0;
		double x = 0.0;
		double x2 = 0.0;
		double datum = 0.0;

		for(int ch=0; ch < vol.getNChannels(); ch++) 
		{
			n = 0;
			x = x2 = 0.0;
			for(int row=0; row < vol.getNCols(); row++) 
			{
				datum = vol.getData(row,col,slice,ch);

				if( datum < 50.0 )
				{
					x += datum;
					x2 += (datum*datum);
					n++;
				}
			}

			if( x == 0 || n == 0 )
			{
				stds[ch] = 1.0;
			}
			else
			{
				stds[ch] = Math.sqrt((double)(x2 - (x*x)/(double)n)/(double)(n-1));
			}
		}
	}

	/** stdAcrossImage - Compute the standard deviation across the 
	 *                whole slice (do it for each channel).
	 */
	protected void stdAcrossImage(final MCVolume vol, final int slice, 
								  double[] stds)
	{
		NoiseEstimate ne = new NoiseEstimate();

		for(int ch=0; ch < vol.getNChannels(); ch++) 
		{
			stds[ch] = (double)ne.gradientMethod( vol.getSlice(slice,ch) );
		}
	}

	/** 
	 *  Set the noise estimation method.
	 */
	public void setNoiseEstimateMethod( final String ne )
	{
		if( ne.indexOf("row") != -1 )
		{
			setNoiseEstimateMethod( NE_ACROSS_ROW );
		}
		else if( ne.indexOf("col") != -1 )
		{
			setNoiseEstimateMethod( NE_ACROSS_COLUMN );
		}
		else if( ne.indexOf("image") != -1 )
		{
			setNoiseEstimateMethod( NE_ACROSS_IMAGE );
		}
		else if( ne.indexOf("none") != -1 )
		{
			setNoiseEstimateMethod( NE_NONE );
		}
		else   
		{
			System.out.println("t2map: Unknown noise estimate type " + ne);
			System.exit(-1);
		}
	}

	/** 
	 *  Set the noise estimation method.
	 */
	public void setNoiseEstimateMethod( final int noise_estimate )
	{
		this.noise_estimate = noise_estimate;
	}
}
